import gevent.monkey
gevent.monkey.patch_all()  # noqa

import boto3
import botocore
import json
import time

from flask import Blueprint
from flask import jsonify
from flask import request
from flask import Response

from app.aws import format_elements
from app.aws import get_amis
from app.aws import get_instance_types
from app.aws import get_nodes
from app.aws import get_rds_instances
from app.aws import get_regions
from app.aws import get_status
from app.aws import limit_elements
from app.aws import sort_by_group
from app.basic_auth import requires_auth
from app.config import Config
from app.log import getLogger
from app.log import getRequestLogger
from app.log import log_request
from app.ssl_279 import _ssl
from app.utils import sorted_json

_ssl  # hack to avoid "Imported but not used" validation issue
blueprint_http = Blueprint('blueprint_http', __name__)
logger = getLogger('haldane')
request_logger = getRequestLogger()
aws_docs_domain = 'http://docs.aws.amazon.com'
error_link = '{0}/AWSEC2/latest/APIReference/errors-overview.html'.format(
    aws_docs_domain)


@blueprint_http.app_errorhandler(404)
def page_not_found(e):
    return Response(
        json.dumps({'status': 'Not Found'}),
        mimetype='application/json',
        status=404
    )


@blueprint_http.errorhandler(boto3.exceptions.Boto3Error)
def handle_boto3_response_error(error):
    logger.info('{0}'.format(error))
    response = jsonify({
        'type': error_link,
        'title': '{0} Error in EC2 Response'.format(error),
        'status': 500,
    })
    response.status_code = 500
    return response


@blueprint_http.errorhandler(botocore.exceptions.BotoCoreError)
def handle_botocore_response_error(error):
    logger.info('{0}'.format(vars(error)))
    response = jsonify({
        'type': error_link,
        'title': '{0} Error in EC2 Response'.format(error),
        'status': 500,
    })
    response.status_code = 500
    return response


@blueprint_http.errorhandler(LookupError)
def handle_lookup_error(error):
    logger.info('LookupError: {0}'.format(error.message))
    response = jsonify({
        'title': 'Invalid argument passed',
        'detail': error.message,
        'status': 400,
    })
    response.status_code = 400
    return response


@blueprint_http.after_request
def log_http_request(response):
    log_request(request, response, request_logger)
    return response


@blueprint_http.route('/')
@blueprint_http.route('/_status')
def status():
    return json_response({
        '_service': {
            '_settings': {
                'regions': Config.AWS_REGIONS,
                'cache': {
                    'expiration': Config.CACHE_EXPIRATION,
                    'size': Config.CACHE_SIZE
                }
            },
            'environment': 'dev' if Config.DEBUG else 'prod',
            'name': 'haldane'
        },
        'meta': {
            'status': 200
        },
        'ok': True
    })


@blueprint_http.route('/amis')
@requires_auth
def amis():
    time_start = time.time()
    query = request.args.get('query', request.args.get('q'))
    regions = get_regions(request.args.get('region'))
    amis = get_amis(request.args, regions, query)

    total_amis = len(amis)
    amis = limit_elements(amis, limit=request.args.get('limit'))
    total_not_hidden = len(amis)
    amis = format_elements(
        amis,
        fields=request.args.get('fields'),
        format=request.args.get('format'))
    total_hidden = total_not_hidden - len(amis)

    if request.args.get('format') == 'csv':
        return csv_response(amis)

    return json_response({
        'meta': {
            'took': time.time() - time_start,
            'total': total_amis,
            'hidden_nodes': total_hidden,
            'regions': regions,
            'per_page': len(amis)
        },
        'amis': amis,
    })


@blueprint_http.route('/instance-types')
@blueprint_http.route('/instance-types/<version>')
@requires_auth
def instance_types(version=None):
    time_start = time.time()
    instance_types, version = get_instance_types(version)

    total_instance_types = len(instance_types)

    return json_response({
        'meta': {
            'api_version': version,
            'took': time.time() - time_start,
            'total': total_instance_types,
            'per_page': len(instance_types)
        },
        'instance-types': instance_types,
    })


@blueprint_http.route('/instances/group')
@blueprint_http.route('/instances/group/<group>')
@blueprint_http.route('/nodes/group')
@blueprint_http.route('/nodes/group/<group>')
@requires_auth
def nodes_by_group(group=None):
    time_start = time.time()
    query = request.args.get('query', request.args.get('q'))
    regions = get_regions(request.args.get('region'))
    status = get_status(request.args.get('status'))
    nodes = get_nodes(request.args,
                      regions,
                      query,
                      status=status)
    groups = sort_by_group(nodes, group=group)

    total_hidden = 0
    _groups = {}
    for group, nodes in groups.items():
        total_nodes = len(nodes)
        _groups[group] = format_elements(
            nodes,
            fields=request.args.get('fields'),
            format=request.args.get('format', 'dict'))
        total_hidden += total_nodes - len(_groups[group])

    return json_response({
        'meta': {
            'took': time.time() - time_start,
            'total': len(groups),
            'hidden_nodes': total_hidden,
            'regions': regions,
            'per_page': len(groups)
        },
        'groups': _groups
    })


@blueprint_http.route('/instances')
@blueprint_http.route('/instances/<region>')
@blueprint_http.route('/nodes')
@blueprint_http.route('/nodes/<region>')
@requires_auth
def nodes(region=None):
    time_start = time.time()
    query = request.args.get('query', request.args.get('q'))
    regions = get_regions(request.args.get('region', region))
    status = get_status(request.args.get('status'))
    nodes = get_nodes(request.args,
                      regions,
                      query,
                      status=status)

    total_nodes = len(nodes)
    nodes = limit_elements(nodes, limit=request.args.get('limit'))
    total_not_hidden = len(nodes)
    nodes = format_elements(
        nodes,
        fields=request.args.get('fields'),
        format=request.args.get('format'))
    total_hidden = total_not_hidden - len(nodes)

    if request.args.get('format') == 'csv':
        return csv_response(nodes)

    return json_response({
        'meta': {
            'took': time.time() - time_start,
            'total': total_nodes,
            'hidden_nodes': total_hidden,
            'regions': regions,
            'per_page': len(nodes)
        },
        'nodes': nodes
    })


@blueprint_http.route('/rds-instances')
@blueprint_http.route('/rds-instances/<region>')
@requires_auth
def rds_instances(region=None):
    time_start = time.time()
    query = request.args.get('query', request.args.get('q'))
    regions = get_regions(request.args.get('region', region))
    status = get_status(request.args.get('status'))
    rds_instances = get_rds_instances(request.args,
                                      regions,
                                      query,
                                      status=status)

    total_rds_instances = len(rds_instances)
    rds_instances = limit_elements(rds_instances,
                                   limit=request.args.get('limit'))
    total_not_hidden = len(rds_instances)
    rds_instances = format_elements(
        rds_instances,
        fields=request.args.get('fields'),
        format=request.args.get('format'))
    total_hidden = total_not_hidden - len(rds_instances)

    if request.args.get('format') == 'csv':
        return csv_response(rds_instances)

    return json_response({
        'meta': {
            'took': time.time() - time_start,
            'total': total_rds_instances,
            'hidden_rds_instances': total_hidden,
            'regions': regions,
            'per_page': len(rds_instances)
        },
        'rds-instances': rds_instances
    })
    pass


def json_response(data):
    return Response(
        sorted_json(data),
        mimetype='application/json',
    )


def csv_response(data):
    resp = []
    header = []
    fields = request.args.get('fields')
    for instance, data in data.items():
        header = data.keys()
        values = data.values()
        if fields:
            values = []
            for field in fields.split(','):
                values.append(str(data.get(field, '')))
        resp.append(','.join(values))

    if fields:
        header = [fields]
    else:
        header = [','.join(header)]

    return Response(
        "\n".join(header + resp),
        mimetype="text/csv"
    )



import inspect
__ssl__ = __import__('ssl')

try:
    _ssl = __ssl__._ssl
except AttributeError:
    _ssl = __ssl__._ssl2


def new_sslwrap(sock,
                server_side=False,
                keyfile=None,
                certfile=None,
                cert_reqs=__ssl__.CERT_NONE,
                ssl_version=__ssl__.PROTOCOL_SSLv23,
                ca_certs=None,
                ciphers=None):
    context = __ssl__.SSLContext(ssl_version)
    context.verify_mode = cert_reqs or __ssl__.CERT_NONE
    if ca_certs:
        context.load_verify_locations(ca_certs)
    if certfile:
        context.load_cert_chain(certfile, keyfile)
    if ciphers:
        context.set_ciphers(ciphers)

    caller_self = inspect.currentframe().f_back.f_locals['self']
    return context._wrap_socket(sock,
                                server_side=server_side,
                                ssl_sock=caller_self)

if not hasattr(_ssl, 'sslwrap'):
    _ssl.sslwrap = new_sslwrap



import os
from app.utils import to_bool


class Config(object):
    ALLOWED_IPS = filter(None, os.getenv('ALLOWED_IPS', '').split(','))
    ALTERNATIVE_AUTOSCALE_TAG_NAME = os.getenv('ALTERNATIVE_AUTOSCALE_TAG_NAME', None)
    AWS_ACCESS_KEY_ID = os.getenv('AWS_ACCESS_KEY_ID')
    AWS_API_VERSION = os.getenv('AWS_API_VERSION', '2016-09-15')
    AWS_REGIONS = filter(None, os.getenv('AWS_REGIONS', 'us-east-1').split(','))
    AWS_SECRET_ACCESS_KEY = os.getenv('AWS_SECRET_ACCESS_KEY')
    BASIC_AUTH = os.getenv('BASIC_AUTH')
    BOOLEAN_AWS_TAG_ATTRIBUTES = filter(None, os.getenv('BOOLEAN_AWS_TAG_ATTRIBUTES', '').split(','))
    BUGSNAG_API_KEY = os.getenv('BUGSNAG_API_KEY')
    CACHE_EXPIRATION = int(os.getenv('CACHE_EXPIRATION', 180))
    CACHE_SIZE = int(os.getenv('CACHE_SIZE', 1024))
    DEBUG = to_bool(os.getenv('DEBUG', 0))
    LISTEN_INTERFACE = os.getenv('LISTEN_INTERFACE', '0.0.0.0')
    PORT = int(os.getenv('PORT', 5000))
    SENTRY_DSN = os.getenv('SENTRY_DSN')
    TOP_LEVEL_AWS_TAG_ATTRIBUTES = filter(None, os.getenv('TOP_LEVEL_AWS_TAG_ATTRIBUTES', '').split(','))



import functools

from flask import request
from flask import Response

from app.config import Config
from app.log import getLogger


def skip_check():
    """This function allows us to skip basic auth check
    for requests coming from specific allowed ips
    """
    logger = getLogger('haldane')
    provided_ips = request.access_route
    return len(provided_ips) > 0 and provided_ips[0] in Config.ALLOWED_IPS


def check_auth(username, password):
    """This function is called to check if a username /
    password combination is valid.
    """
    authentications = Config.BASIC_AUTH.split(',')
    authentication = '{0}:{1}'.format(username, password)

    return authentication in authentications


def authenticate():
    """Sends a 401 response that enables basic auth"""
    return Response(
        'Could not verify your access level for that URL.\n'
        'You have to login with proper credentials', 401,
        {'WWW-Authenticate': 'Basic realm="Login Required"'})


def requires_auth(f):
    @functools.wraps(f)
    def decorated(*args, **kwargs):
        if skip_check() or not Config.BASIC_AUTH:
            return f(*args, **kwargs)

        auth = request.authorization
        if not auth or not check_auth(auth.username, auth.password):
            return authenticate()
        return f(*args, **kwargs)
    return decorated



# -*- coding: utf-8 -*-
"""
log:
    Encapsulates all of the log setup for seatgeek default loggers.

    Environment Variables:
        LOGGING_VERBOSE: Should run logger in verbose mode, which outputs
            logger.debug() messages. Default = False.

    Interface:
        from sglib import log
        logger = log.getLogger("my.logger")
        logger.info("Hi There)
        >> [2012-11-12 15:05:22] my.logger - INFO - Hi There
"""
import datetime
import os
import sys
import logging
import logging.config

from werkzeug._internal import _log
from werkzeug.serving import WSGIRequestHandler


def _to_bool(s):
    try:
        int_s = int(s)
        return bool(int_s) is True
    except:
        pass

    if isinstance(s, basestring):
        if s.lower() in ['true', 't']:
            return True
        if s.lower() in ['false', 'f']:
            return False

    return bool(s) is True


# log settings for root logger by default
LOG_SETTINGS = {
    'version': 1,
    'disable_existing_loggers': True,
    'formatters': {
        'normal': {
            'format': "[%(asctime)s] %(name)s [pid:%(process)d] - %(levelname)s - %(message)s",
            'datefmt': "%Y-%m-%d %H:%M:%S"
        },
        'werkzeug': {
            'format': "[%(asctime)s] %(name)s [pid:%(process)d] - %(levelname)s - %(message)s",
            'datefmt': "%Y-%m-%d %H:%M:%S"
        },
    },
    'handlers': {
        'verbose': {
            'level': 'DEBUG',
            'class': 'logging.StreamHandler',
            'formatter': 'normal',
            'stream': 'ext://sys.stdout'
        },
        'console': {
            'level': 'INFO',
            'class': 'logging.StreamHandler',
            'formatter': 'normal',
            'stream': 'ext://sys.stdout'
        },
        'console_werkzeug': {
            'level': 'INFO',
            'class': 'logging.StreamHandler',
            'formatter': 'werkzeug',
            'stream': 'ext://sys.stdout'
        },
        'error': {
            'level': 'ERROR',
            'class': 'logging.StreamHandler',
            'formatter': 'normal',
            'stream': 'ext://sys.stderr'
        },
    },
    'loggers': {
        'sqlalchemy.engine': {
            'propagate': True
        },
        'werkzeug': {
            'handlers': ['console_werkzeug'],
            'level': 'DEBUG',
            'propagate': False,
        },
    },
    'root': {
        'level': 'NOTSET',
        'handlers': ['console', 'error']
    },
}

normal_formatter = logging.Formatter(
    fmt=LOG_SETTINGS['formatters']['normal']['format'],
    datefmt=LOG_SETTINGS['formatters']['normal']['datefmt'])

werkzeug_formatter = logging.Formatter(
    fmt=LOG_SETTINGS['formatters']['werkzeug']['format'],
    datefmt=LOG_SETTINGS['formatters']['werkzeug']['datefmt'])

werkzeug_stream_handler = logging.StreamHandler(sys.stdout)
werkzeug_stream_handler.setFormatter(werkzeug_formatter)

werkzeug_error_stream_handler = logging.StreamHandler(sys.stderr)
werkzeug_error_stream_handler.setFormatter(werkzeug_formatter)
werkzeug_error_stream_handler.setLevel(logging.ERROR)

stream_handler = logging.StreamHandler(sys.stdout)
stream_handler.setFormatter(normal_formatter)

error_stream_handler = logging.StreamHandler(sys.stderr)
error_stream_handler.setFormatter(normal_formatter)
error_stream_handler.setLevel(logging.ERROR)

# environment flags we should honor
BUGSNAG_API_KEY = os.getenv('BUGSNAG_API_KEY')
LOGGING_VERBOSE = _to_bool(os.getenv('LOGGING_VERBOSE', False))
SENTRY_DSN = os.getenv('SENTRY_DSN', None)

if LOGGING_VERBOSE:
    LOG_SETTINGS['root']['handlers'].remove('console')
    LOG_SETTINGS['root']['handlers'].append('verbose')

logging.config.dictConfig(LOG_SETTINGS)

if BUGSNAG_API_KEY:
    import bugsnag

if SENTRY_DSN:
    from raven import Client as RavenClient
    from raven.conf import setup_logging
    from raven.handlers.logging import SentryHandler
    assert setup_logging


def _configure_error_handler():
    handler = None
    if BUGSNAG_API_KEY:
        client = bugsnag.Client(api_key=BUGSNAG_API_KEY)
        handler = client.log_handler()
        handler.setLevel(logging.ERROR)

    if SENTRY_DSN:
        client = RavenClient(SENTRY_DSN)
        handler = SentryHandler(client)
        handler.setLevel(logging.ERROR)
    return handler


def getLogger(name, verbose=LOGGING_VERBOSE):
    """
    returns a python logger object configured according to seatgeek's standards

    Args:
        verbose: [True/False] -- default is set by env variable LOGGING_VERBOSE
    """
    logger = logging.getLogger(name)

    # just in case that was not a new logger, get rid of all the handlers
    # already attached to it.
    del logger.handlers[:]

    # handle verbosity
    if verbose:
        logger.setLevel(logging.DEBUG)
        logger.addHandler(stream_handler)
        logger.addHandler(error_stream_handler)
    else:
        logger.setLevel(logging.INFO)
        logger.addHandler(stream_handler)
        logger.addHandler(error_stream_handler)

    logger.propagate = False
    error_handler = _configure_error_handler()
    if error_handler:
        logger.addHandler(error_handler)

    return logger


def getRequestLogger(verbose=LOGGING_VERBOSE):
    logger = logging.getLogger('werkzeug')

    # just in case that was not a new logger, get rid of all the handlers
    # already attached to it.
    del logger.handlers[:]

    # handle verbosity
    if verbose:
        logger.setLevel(logging.DEBUG)
        logger.addHandler(werkzeug_stream_handler)
        logger.addHandler(werkzeug_error_stream_handler)
    else:
        logger.setLevel(logging.INFO)
        logger.addHandler(werkzeug_stream_handler)
        logger.addHandler(werkzeug_error_stream_handler)

    logger.propagate = False
    error_handler = _configure_error_handler()
    if error_handler:
        logger.addHandler(error_handler)

    return logger


def log_request(request, response, logger=None):
    server_software = request.environ.get('SERVER_SOFTWARE')
    if logger is None or 'gunicorn' not in server_software:
        return

    now = datetime.datetime.utcnow()
    message = 'remote-addr={0} method={1} path={2} protocol={3} status_code={4}'
    logger.info(message.format(
        request.environ.get("REMOTE_ADDR"),
        request.method,
        request.path,
        request.environ.get("SERVER_PROTOCOL"),
        response.status_code,
    ))


def werkzeug_log(self, type, message, *args):
    _log(type, 'remote-addr={0} {1}\n'.format(
         self.address_string(),
         message % args))


def werkzeug_log_request(self, code='-', size='-'):
    self.log('info', 'method={0} path={1} protocol={2} status_code={3}'.format(
        self.command,
        self.path,
        self.request_version,
        code))


WSGIRequestHandler.log = werkzeug_log
WSGIRequestHandler.log_request = werkzeug_log_request



from app.config import Config
from app.utils import to_bool


def filter_elements(elements, request_args, query=None, status=None):
    request_args = request_args.to_dict()
    if query:
        request_args['substring.name'] = query
    elements = filter_by_args(elements, request_args)
    elements = filter_by_tags(elements, request_args)

    return elements


def filter_by_args(elements, args):
    valid_search_keys = [
        # common
        'availability_zone',
        'id',
        'name',
        'region',
        'status',

        #  ec2
        'elastic_ip',
        'group',
        'image_id',
        'image_name',
        'instance_type',
        'instance_class',
        'instance_profile_id',
        'instance_profile_name',
        'ip_address',
        'private_ip_address',
        'vpc_id',

        # rds
        'allocated_storage',
        'auto_minor_version_upgrade',
        'backup_retention_period',
        'ca_certificate_identifier',
        'copy_tags_to_snapshot',
        'db_instance_arn',
        'db_instance_class',
        'db_instance_port',
        'db_instance_status',
        'db_name',
        'dbi_resource_id',
        'engine',
        'engine_version',
        'enhanced_monitoring_resource_arn',
        'license_model',
        'master_username',
        'monitoring_interval',
        'monitoring_role_arn',
        'multi_az',
        'preferred_backup_window',
        'preferred_maintenance_window',
        'publicly_accessible',
        'secondary_availability_zone',
        'storage_encrypted',
        'storage_type',
    ] + Config.BOOLEAN_AWS_TAG_ATTRIBUTES
    bool_search_keys = [
        'elastic_ip',
    ] + Config.BOOLEAN_AWS_TAG_ATTRIBUTES
    not_starts_with_filters, args = get_filter(args, 'not-starts-with', valid_search_keys=valid_search_keys, bool_search_keys=bool_search_keys)
    not_ends_with_filters, args = get_filter(args, 'not-ends-with', valid_search_keys=valid_search_keys, bool_search_keys=bool_search_keys)
    not_substring_filters, args = get_filter(args, 'not-substring', valid_search_keys=valid_search_keys, bool_search_keys=bool_search_keys)
    not_in_list_filters, args = get_filter(args, 'not-in-list', valid_search_keys=valid_search_keys, bool_search_keys=bool_search_keys)
    starts_with_filters, args = get_filter(args, 'starts-with', valid_search_keys=valid_search_keys, bool_search_keys=bool_search_keys)
    ends_with_filters, args = get_filter(args, 'ends-with', valid_search_keys=valid_search_keys, bool_search_keys=bool_search_keys)
    substring_filters, args = get_filter(args, 'substring', valid_search_keys=valid_search_keys, bool_search_keys=bool_search_keys)
    is_false_filters, args = get_filter(args, 'is-false', valid_search_keys=valid_search_keys, bool_search_keys=bool_search_keys)
    is_true_filters, args = get_filter(args, 'is-true', valid_search_keys=valid_search_keys, bool_search_keys=bool_search_keys)
    is_null_filters, args = get_filter(args, 'is-null', valid_search_keys=valid_search_keys, bool_search_keys=bool_search_keys)
    in_list_filters, args = get_filter(args, 'in-list', valid_search_keys=valid_search_keys, bool_search_keys=bool_search_keys)
    exact_filters, args = get_filter(args, 'exact', valid_search_keys=valid_search_keys, bool_search_keys=bool_search_keys)

    bool_search_keys = [
        'elastic_ip',
    ]

    for key, value in exact_filters.items():
        _elements = []
        for element in elements:
            attribute = element.get(key, '')
            if not attribute and value:
                continue

            if key in bool_search_keys:
                attribute = to_bool(attribute)

            if value == attribute:
                _elements.append(element)
        elements = _elements
    for key, value in is_null_filters.items():
        _elements = []
        for element in elements:
            attribute = element.get(key, '')
            if attribute is None:
                _elements.append(element)
        elements = _elements
    for key, value in is_true_filters.items():
        _elements = []
        for element in elements:
            attribute = element.get(key, None)
            if attribute is True:
                _elements.append(element)
        elements = _elements
    for key, value in is_false_filters.items():
        _elements = []
        for element in elements:
            attribute = element.get(key, None)
            if attribute is False:
                _elements.append(element)
        elements = _elements
    for key, value in in_list_filters.items():
        _elements = []
        for element in elements:
            attribute = element.get(key)
            if not attribute:
                continue

            if attribute in value.split(','):
                _elements.append(element)
        elements = _elements
    for key, value in substring_filters.items():
        _elements = []
        for element in elements:
            attribute = element.get(key)
            if not attribute:
                continue

            if value in attribute:
                _elements.append(element)
        elements = _elements
    for key, value in starts_with_filters.items():
        _elements = []
        for element in elements:
            attribute = element.get(key)
            if not attribute:
                continue

            if attribute.startswith(value):
                _elements.append(element)
        elements = _elements
    for key, value in ends_with_filters.items():
        _elements = []
        for element in elements:
            attribute = element.get(key)
            if not attribute:
                continue

            if attribute.endswith(value):
                _elements.append(element)
        elements = _elements

    for key, value in not_in_list_filters.items():
        _elements = []
        for element in elements:
            attribute = element.get(key)
            if not attribute:
                continue

            if attribute not in value.split(','):
                _elements.append(element)
        elements = _elements
    for key, value in not_substring_filters.items():
        _elements = []
        for element in elements:
            attribute = element.get(key, None)
            if attribute is None:
                continue

            if value not in attribute:
                _elements.append(element)
        elements = _elements
    for key, value in not_starts_with_filters.items():
        _elements = []
        for element in elements:
            attribute = element.get(key)
            if not attribute:
                continue

            if not attribute.startswith(value):
                _elements.append(element)
        elements = _elements
    for key, value in not_ends_with_filters.items():
        _elements = []
        for element in elements:
            attribute = element.get(key)
            if not attribute:
                continue

            if not attribute.endswith(value):
                _elements.append(element)
        elements = _elements

    return elements


def filter_by_tags(elements, args):
    not_starts_with_filters, args = get_filter(args, 'tags.not-starts-with')
    not_ends_with_filters, args = get_filter(args, 'tags.not-ends-with')
    not_substring_filters, args = get_filter(args, 'tags.not-substring')
    not_in_list_filters, args = get_filter(args, 'tags.not-in-list')
    starts_with_filters, args = get_filter(args, 'tags.starts-with')
    ends_with_filters, args = get_filter(args, 'tags.ends-with')
    substring_filters, args = get_filter(args, 'tags.substring')
    is_false_filters, args = get_filter(args, 'tags.is-false')
    is_true_filters, args = get_filter(args, 'tags.is-true')
    is_null_filters, args = get_filter(args, 'tags.is-null')
    in_list_filters, args = get_filter(args, 'tags.in-list')
    exact_filters, args = get_filter(args, 'tags.exact')

    for key, value in exact_filters.items():
        _elements = []
        for element in elements:
            attribute = element.get('tags', {}).get(key)
            if not attribute and value:
                continue

            if value == attribute:
                _elements.append(element)
        elements = _elements
    for key, value in is_null_filters.items():
        _elements = []
        for element in elements:
            attribute = element.get('tags', {}).get(key, '')
            if attribute is None:
                _elements.append(element)
        elements = _elements
    for key, value in is_true_filters.items():
        _elements = []
        for element in elements:
            attribute = element.get('tags', {}).get(key, None)
            if attribute is True:
                _elements.append(element)
        elements = _elements
    for key, value in is_false_filters.items():
        _elements = []
        for element in elements:
            attribute = element.get('tags', {}).get(key, None)
            if attribute is False:
                _elements.append(element)
        elements = _elements
    for key, value in in_list_filters.items():
        _elements = []
        for element in elements:
            attribute = element.get('tags', {}).get(key)
            if not attribute:
                continue

            if value in attribute.split(','):
                _elements.append(element)
        elements = _elements
    for key, value in substring_filters.items():
        _elements = []
        for element in elements:
            attribute = element.get('tags', {}).get(key)
            if not attribute:
                continue

            if value in attribute:
                _elements.append(element)
        elements = _elements
    for key, value in starts_with_filters.items():
        _elements = []
        for element in elements:
            attribute = element.get('tags', {}).get(key)
            if not attribute:
                continue

            if attribute.startswith(value):
                _elements.append(element)
        elements = _elements
    for key, value in ends_with_filters.items():
        _elements = []
        for element in elements:
            attribute = element.get('tags', {}).get(key)
            if not attribute:
                continue

            if attribute.endswith(value):
                _elements.append(element)
        elements = _elements

    for key, value in not_in_list_filters.items():
        _elements = []
        for element in elements:
            attribute = element.get('tags', {}).get(key)
            if not attribute:
                continue

            if value not in attribute.split(','):
                _elements.append(element)
        elements = _elements
    for key, value in not_substring_filters.items():
        _elements = []
        for element in elements:
            attribute = element.get('tags', {}).get(key, None)
            if attribute is None:
                continue

            if value not in attribute:
                _elements.append(element)
        elements = _elements
    for key, value in not_starts_with_filters.items():
        _elements = []
        for element in elements:
            attribute = element.get('tags', {}).get(key)
            if not attribute:
                continue

            if not attribute.startswith(value):
                _elements.append(element)
        elements = _elements
    for key, value in not_ends_with_filters.items():
        _elements = []
        for element in elements:
            attribute = element.get('tags', {}).get(key)
            if not attribute:
                continue

            if not attribute.endswith(value):
                _elements.append(element)
        elements = _elements

    return elements


def get_filter(args, filter_name, filter_key_prefix=None, valid_search_keys=None, bool_search_keys=None):
    filters = {}
    unused_args = {}
    for key, value in args.items():
        if bool_search_keys and key in bool_search_keys:
            value = to_bool(value)

        if filter_name == 'tags.exact' and key.startswith('tags.') and not key.startswith('tags.exact.'):
            key = 'tags.exact.{0}'.format(key.replace('tags.', ''))
        elif filter_name == 'exact' and not key.startswith('exact.') and '.' not in key:
            if valid_search_keys and key.replace('exact.', '') not in valid_search_keys:
                unused_args[key] = value
                continue
            key = 'exact.{0}'.format(key.replace('exact.', ''))
        elif not key.startswith('{0}.'.format(filter_name)):
            unused_args[key] = value
            continue

        length = len(filter_name) + 1
        filter_key_lookup = key[length:]
        filter_key = filter_key_lookup

        if valid_search_keys and filter_key_lookup in valid_search_keys:
            filters[filter_key] = value
        elif not valid_search_keys:
            filters[filter_key] = value

    return filters, unused_args



from gunicorn.glogging import Logger as GunicornLogger


class Logger(GunicornLogger):
    error_fmt = r"[%(asctime)s] %(name)s [pid:%(process)d] - %(levelname)s - %(message)s"
    datefmt = r"%Y-%m-%d %H:%M:%S"

    access_fmt = "%(message)s"
    syslog_fmt = "[%(process)d] %(message)s"



import gevent.monkey
gevent.monkey.patch_all()  # noqa

import boto3
import datetime
import json
import lru

from botocore import BOTOCORE_ROOT
from inflection import underscore

from app.config import Config
from app.filters import filter_elements
from app.log import getLogger
from app.utils import sorted_dict
from app.utils import to_bool


logger = getLogger('haldane')


def get_amis(request_args, regions, query=None):
    amis = []
    for region in regions:
        ec2_resource = boto3.resource('ec2', region_name=region)
        amis.extend(get_amis_in_region(ec2_resource, region))

    return filter_elements(amis, request_args, query=query)


def get_nodes(request_args, regions, query=None, status=None):
    nodes = []
    for region in regions:
        nodes.extend(get_nodes_in_region(region))

    return filter_elements(nodes, request_args, query=query, status=status)


def get_rds_instances(request_args, regions, query=None, status=None):
    rds_instances = []
    for region in regions:
        rds_instances.extend(get_rds_instances_in_region(region))

    return filter_elements(rds_instances,
                           request_args,
                           query=query,
                           status=status)


def get_regions(region=None):
    regions = []
    if region is None:
        regions = Config.AWS_REGIONS
    else:
        if region not in Config.AWS_REGIONS:
            raise LookupError('Invalid region querystring argument passed')
        regions = [region]
    return regions


@lru.lru_cache_function(
    max_size=Config.CACHE_SIZE,
    expiration=Config.CACHE_EXPIRATION)
def get_amis_in_region(resource, region):
    amis = []
    for image in resource.images.filter(Owners=['self']).all():
        ami = {
            'architecture': image.architecture,
            'creation_date': image.creation_date,
            'description': image.description,
            'hypervisor': image.hypervisor,
            'id': image.image_id,
            'is_public': image.public,
            'kernel_id': image.kernel_id,
            'location': image.image_location,
            'name': image.name,
            'owner_alias': image.image_owner_alias,
            'owner_id': image.owner_id,
            'platform': image.platform,
            'product_codes': image.product_codes,
            'ramdisk_id': image.ramdisk_id,
            'region': region,
            'root_device_name': image.root_device_name,
            'root_device_type': image.root_device_type,
            'sriov_net_support': image.sriov_net_support,
            'state': image.state,
            'type': image.image_type,
            'virtualization_type': image.virtualization_type,
        }

        block_device_fields = [
            'attach_time',
            'delete_on_termination',
            'ebs',
            'encrypted',
            'ephemeral_name',
            'iops',
            'no_device',
            'snapshot_id',
            'status',
            'volume_id',
            'volume_size',
            'volume_type',
        ]

        block_device_mapping = {}
        for block_device in image.block_device_mappings:
            device_name = block_device['DeviceName']
            data = block_device.get('Ebs', {})
            data['ephemeral_name'] = block_device.get('VirtualName')
            data['no_device'] = block_device.get('NoDevice')
            data = dict((underscore(k), v) for k, v in data.iteritems())
            for field in block_device_fields:
                data[field] = data.get(field)
            block_device_mapping[device_name] = data

        tags = image.tags
        if not tags:
            tags = []

        ami['block_device_mapping'] = block_device_mapping
        ami['tags'] = dict((tag['Key'], tag['Value']) for tag in tags)
        amis.append(ami)

    return amis


@lru.lru_cache_function(
    max_size=Config.CACHE_SIZE,
    expiration=Config.CACHE_EXPIRATION)
def get_elastic_ips(ec2_resource):
    classic_addresses = ec2_resource.classic_addresses.all()
    return [address.public_ip for address in classic_addresses]


@lru.lru_cache_function(
    max_size=Config.CACHE_SIZE,
    expiration=Config.CACHE_EXPIRATION)
def get_instance_types(version=None):
    if version is None:
        version = Config.AWS_API_VERSION
    filename = '{0}/data/ec2/{1}/service-2.json'.format(BOTOCORE_ROOT, version)
    with open(filename) as f:
        data = json.loads(f.read())
        return data['shapes']['InstanceType']['enum'], version
    return [], version


@lru.lru_cache_function(
    max_size=Config.CACHE_SIZE,
    expiration=Config.CACHE_EXPIRATION)
def get_nodes_in_region(region):
    ec2_resource = boto3.resource('ec2', region_name=region)
    elastic_ips = get_elastic_ips(ec2_resource)
    images = get_amis_in_region(ec2_resource, region)

    instances = []
    instances_for_region = ec2_resource.instances.all()
    for instance in instances_for_region:
        ip_address = instance.public_ip_address
        pip_address = instance.private_ip_address
        instance_id = instance.instance_id

        if ip_address is None:
            ip_address = ''
        if pip_address is None:
            pip_address = ''
        if instance_id is None:
            instance_id = ''

        tags = instance.tags
        if not tags:
            tags = []
        tags = dict((tag['Key'], tag['Value']) for tag in tags)

        for tag, value in tags.items():
            if value.lower() in ['true', 'false']:
                tags[tag] = to_bool(value)
            if value.lower() in ['none', 'nil', 'null']:
                tags[tag] = None
        name = tags.get('Name')

        default_group_name = None
        if Config.ALTERNATIVE_AUTOSCALE_TAG_NAME:
            default_group_name = tags.get(Config.ALTERNATIVE_AUTOSCALE_TAG_NAME, '')
        group = tags.get('aws:autoscaling:groupName', default_group_name)

        if group == '' or group is None:
            group = None
            if not name:
                name = instance_id
        elif not name:
            name = group + '-' + instance_id

        instance_name = name.replace('_', '-').strip()

        image_name = [image['name'] for image in images if image['id'] == instance.image_id]
        if len(image_name) == 1:
            image_name = image_name[0]
        else:
            image_name = ''

        instance_profile_id = None
        instance_profile_name = None
        if instance.iam_instance_profile:
            instance_profile_id = instance.iam_instance_profile['Id']
            instance_profile_name = instance.iam_instance_profile['Arn'].split('/', 1)[1]

        data = {
            'availability_zone': instance.placement['AvailabilityZone'],
            'group': group,
            'elastic_ip': ip_address in elastic_ips,
            'id': instance_id,
            'image_id': instance.image_id,
            'image_name': image_name,
            'instance_type': instance.instance_type,
            'instance_class': instance.instance_type.split('.')[0],
            'ip_address': ip_address,
            'instance_profile_id': instance_profile_id,
            'instance_profile_name': instance_profile_name,
            'launch_time': instance.launch_time.strftime('%Y-%m-%dT%H:%M:%S.000Z'),
            'name': instance_name,
            'private_ip_address': pip_address,
            'region': region,
            'status': instance.state['Name'],
            'tags': tags,
            'vpc_id': instance.vpc_id,
        }

        for key in Config.BOOLEAN_AWS_TAG_ATTRIBUTES:
            data[key] = to_bool(tags.get(key, ''))
        for key in Config.TOP_LEVEL_AWS_TAG_ATTRIBUTES:
            data[key] = tags.get(key, '')

        instances.append(sorted_dict(data))

    return instances


@lru.lru_cache_function(
    max_size=Config.CACHE_SIZE,
    expiration=Config.CACHE_EXPIRATION)
def get_rds_instances_in_region(region):
    client = boto3.client('rds', region_name=region)
    paginator = client.get_paginator('describe_db_instances')
    page_iterator = paginator.paginate()

    rds_instances = []
    for page in page_iterator:
        for data in page['DBInstances']:
            rds_instance = {}
            rds_instance['id'] = data['DbiResourceId']
            rds_instance['name'] = data['DBInstanceIdentifier']
            rds_instance['region'] = region
            rds_instance['status'] = data['DBInstanceStatus']
            del data['DBInstanceIdentifier']
            del data['DBInstanceStatus']
            del data['DbiResourceId']

            for key, value in data.items():
                rds_instance[underscore(key)] = transform_value(value)

            response = client.list_tags_for_resource(ResourceName=rds_instance['db_instance_arn'])
            rds_instance['tags'] = dict((tag['Key'], tag['Value']) for tag in response['TagList'])

            rds_instances.append(rds_instance)

    return rds_instances


def format_elements(elements, fields=None, format=None):
    if fields:
        fields = fields.split(',')
        _elements = []
        for element in elements:
            _element = {}
            if format != 'list':
                _element['id'] = element['id']
                _element['name'] = element['name']
            for field in fields:
                if field in element:
                    _element[field] = element.get(field)
            _elements.append(_element)
        elements = _elements
    if format == 'list':
        return elements

    if fields is None:
        fields = []

    _elements = {}
    for element in elements:
        name = element['name']
        if name in _elements:
            existing = _elements[name]
            logger.warning(u'duplicate name collision: {0}'.format(
                name))
            logger.warning('existing running instance id: {0}'.format(
                existing['id']))
            logger.warning('new instance id: {0}'.format(element['id']))
            if element.get('status', '') == 'running':
                logger.warning('hiding existing resource {0} in state {1}'.format(  # noqa
                    existing['id'],
                    existing.get('status', '')))
            else:
                logger.warning('hiding new resource {0} in state {1}'.format(
                    element['id'],
                    element.get('status', '')))
                continue

        _elements[name] = element

    elements = _elements
    _elements = {}
    for name, element in elements.items():
        if fields and 'name' not in fields:
            del element['name']
        if fields and 'id' not in fields:
            del element['id']
        _elements[name] = element

    return sorted_dict(_elements)


def limit_elements(elements, limit=None):
    if limit is None:
        return elements

    count = 0
    limit = int(limit)
    _elements = []
    for element in elements:
        if count == limit:
            break

        count += 1
        _elements.append(element)
    return _elements


def sort_by_group(nodes, group=None):
    groups = {
        'None': []
    }
    for node in nodes:
        if node.get('group') is None:
            groups['None'].append(node)
        else:
            if node.get('group') not in groups:
                groups[node.get('group')] = []
            groups[node.get('group')].append(node)

    if group is not None:
        if group in groups:
            groups = {group: groups[group]}
        else:
            groups = {}

    return groups


def transform_value(data):
    if type(data) == datetime.datetime:
        data = data.strftime('%Y-%m-%dT%H:%M:%S.000Z')
    elif type(data) == dict:
        value = {}
        for k, v in data.items():
            value[underscore(k)] = transform_value(v)
        data = value
    elif type(data) == list:
        value = []
        for v in data:
            value.append(transform_value(v))
        data = value
    return data


def get_status(status=None):
    if status is None:
        return None
    valid_states = [
        'pending',
        'running',
        'shutting-down',
        'terminated',
        'stopping',
        'stopped',
    ]
    if status not in valid_states:
        raise LookupError('Invalid status querystring argument passed.')

    return status



import collections
import json


def sorted_dict(data):
    return collections.OrderedDict(sorted(
        data.items(), key=lambda t: t[0]))


def sorted_json(data):
    return json.dumps(data, sort_keys=True, indent=2, separators=(',', ': ')),


def to_bool(s):
    try:
        int_s = int(s)
        return bool(int_s) is True
    except:
        pass

    if isinstance(s, basestring):
        if s.lower() in ['true', 't']:
            return True
        if s.lower() in ['false', 'f']:
            return False

    return bool(s) is True



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (c) 2014 SeatGeek

# This file is part of haldane.

from app import __version__
import os

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup


def open_file(fname):
    return open(os.path.join(os.path.dirname(__file__), fname))


def run_setup():
    setup(
        name='haldane',
        version=__version__,
        description='A friendly http interface to the aws api',
        keywords='aws api',
        url='https://github.com/seatgeek/haldane',
        author='SeatGeek',
        author_email='opensource@seatgeek.com',
        license='BSD',
        packages=['app'],
        install_requires=open_file('requirements.txt').readlines(),
        test_suite='tests',
        long_description=open_file('README.rst').read(),
        include_package_data=True,
        zip_safe=True,
        classifiers=[
        ],
    )

if __name__ == '__main__':
    run_setup()
